import importlib
import time
from typing import Iterable, Any  # noqa: F401

from pyhocon import ConfigTree  # noqa: F401

from databuilder.extractor.generic_extractor import GenericExtractor


class Neo4jEsLastUpdatedExtractor(GenericExtractor):
    """
    Extractor to extract last updated timestamp for neo4j and Es
    """
    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        Receives a list of dictionaries which is used for extraction
        :param conf:
        :return:
        """
        self.conf = conf

        model_class = conf.get('model_class', None)
        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)
            last_updated_timestamp = int(time.time())
            result = {'timestamp': last_updated_timestamp}
            results = [self.model_class(**result)]
            self._iter = iter(results)
        else:
            raise RuntimeError('model class needs to be provided!')

    def extract(self):
        # type: () -> Any
        """
        Fetch one sql result row, convert to {model_class} if specified before
        returning.
        :return:
        """
        try:
            result = next(self._iter)
            return result
        except StopIteration:
            return None

    def get_scope(self):
        # type: () -> str
        return 'extractor.neo4j_es_last_updated'



import os

from pyhocon import ConfigTree  # noqa: F401

from databuilder.loader.base_loader import Loader
from databuilder.models.elasticsearch_document import ElasticsearchDocument


class FSElasticsearchJSONLoader(Loader):
    """
    Loader class to produce Elasticsearch bulk load file to Local FileSystem
    """
    FILE_PATH_CONFIG_KEY = 'file_path'
    FILE_MODE_CONFIG_KEY = 'mode'

    def init(self, conf):
        # type: (ConfigTree) -> None
        """

        :param conf:
        :return:
        """
        self.conf = conf
        self.file_path = self.conf.get_string(FSElasticsearchJSONLoader.FILE_PATH_CONFIG_KEY)
        self.file_mode = self.conf.get_string(FSElasticsearchJSONLoader.FILE_MODE_CONFIG_KEY, 'w')

        file_dir = self.file_path.rsplit('/', 1)[0]
        self._ensure_directory_exists(file_dir)
        self.file_handler = open(self.file_path, self.file_mode)

    def _ensure_directory_exists(self, path):
        # type: (str) -> None
        """
        Check to ensure file directory exists; create the directories otherwise
        :param path:
        :return: None
        """
        if os.path.exists(path):
            return  # nothing to do here

        os.makedirs(path)

    def load(self, record):
        # type: (ElasticsearchDocument) -> None
        """
        Write a record in json format to file
        :param record:
        :return:
        """
        if not record:
            return

        if not isinstance(record, ElasticsearchDocument):
            raise Exception("Record not of type 'ElasticsearchDocument'!")

        self.file_handler.write(record.to_json())
        self.file_handler.flush()

    def close(self):
        # type: () -> None
        """
        close the file handler
        :return:
        """
        self.file_handler.close()

    def get_scope(self):
        # type: () -> str
        return 'loader.filesystem.elasticsearch'



import logging

from pyhocon import ConfigTree  # noqa: F401

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor  # noqa: F401
from databuilder.loader.base_loader import Loader  # noqa: F401
from databuilder.task.base_task import Task  # noqa: F401
from databuilder.transformer.base_transformer import Transformer  # noqa: F401
from databuilder.transformer.base_transformer \
    import NoopTransformer  # noqa: F401
from databuilder.utils.closer import Closer


class DefaultTask(Task):
    """
    A default task expecting to extract, transform and load.

    """
    def __init__(self,
                 extractor,
                 loader,
                 transformer=NoopTransformer()):
        # type: (Extractor, Loader, Transformer) -> None
        self.extractor = extractor
        self.transformer = transformer
        self.loader = loader

        self._closer = Closer()
        self._closer.register(self.extractor.close)
        self._closer.register(self.transformer.close)
        self._closer.register(self.loader.close)

    def init(self, conf):
        # type: (ConfigTree) -> None
        self.extractor.init(Scoped.get_scoped_conf(conf, self.extractor.get_scope()))
        self.transformer.init(Scoped.get_scoped_conf(conf, self.transformer.get_scope()))
        self.loader.init(Scoped.get_scoped_conf(conf, self.loader.get_scope()))

    def run(self):
        # type: () -> None
        """
        Runs a task
        :return:
        """
        logging.info('Running a task')
        try:
            record = self.extractor.extract()

            while record:
                record = self.transformer.transform(record)
                if not record:
                    continue
                self.loader.load(record)
                record = self.extractor.extract()
        finally:
            self._closer.close()



from typing import List, Optional  # noqa: F401


class Neo4jDataResult:
    """
    Neo4j Graph data model
    CYPHER QUERY returns one column per row
    """
    def __init__(self,
                 database,  # type: str
                 cluster,  # type: str
                 schema_name,  # type: str
                 table_name,  # type: str
                 table_key,  # type: str
                 table_description,  # type: str
                 table_last_updated_epoch,  # type: Optional[int]
                 column_names,  # type: List[str]
                 column_descriptions,  # type: List[str]
                 total_usage,  # type: int
                 unique_usage,  # type: int
                 tag_names,  # type: List[str]
                 ):
        # type: (...) -> None
        self.database = database
        self.cluster = cluster
        self.schema_name = schema_name
        self.table_name = table_name
        self.table_key = table_key
        self.table_description = table_description
        self.table_last_updated_epoch = int(table_last_updated_epoch) if table_last_updated_epoch else None
        self.column_names = column_names
        self.column_descriptions = column_descriptions
        self.total_usage = total_usage
        self.unique_usage = unique_usage
        self.tag_names = tag_names



import abc

from pyhocon import ConfigTree  # noqa: F401

from databuilder import Scoped


class Task(Scoped):
    """
    A Abstract task that can run an abstract task
    """
    @abc.abstractmethod
    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    @abc.abstractmethod
    def run(self):
        # type: () -> None
        """
        Runs a task
        :return:
        """
        pass

    def get_scope(self):
        # type: () -> str
        return 'task'



import logging

from pyhocon import ConfigTree  # noqa: F401
from statsd import StatsClient

from databuilder import Scoped
from databuilder.job.base_job import Job
from databuilder.publisher.base_publisher import NoopPublisher
from databuilder.publisher.base_publisher import Publisher  # noqa: F401
from databuilder.task.base_task import Task  # noqa: F401

LOGGER = logging.getLogger(__name__)


class DefaultJob(Job):
    # Config keys
    IS_STATSD_ENABLED = 'is_statsd_enabled'
    JOB_IDENTIFIER = 'identifier'

    """
    Default job that expects a task, and optional publisher
    If configured job will emit success/fail metric counter through statsd where prefix will be
    amundsen.databuilder.job.[identifier] .
    Note that job.identifier is part of metrics prefix and choose unique & readable identifier for the job.

    To configure statsd itself, use environment variable: https://statsd.readthedocs.io/en/v3.2.1/configure.html
    """
    def __init__(self,
                 conf,
                 task,
                 publisher=NoopPublisher()):
        # type: (Task, ConfigTree, Publisher) -> None
        self.task = task
        self.conf = conf
        self.publisher = publisher
        self.scoped_conf = Scoped.get_scoped_conf(self.conf,
                                                  self.get_scope())
        if self.scoped_conf.get_bool(DefaultJob.IS_STATSD_ENABLED, False):
            prefix = 'amundsen.databuilder.job.{}'.format(self.scoped_conf.get_string(DefaultJob.JOB_IDENTIFIER))
            LOGGER.info('Setting statsd for job metrics with prefix: {}'.format(prefix))
            self.statsd = StatsClient(prefix=prefix)
        else:
            self.statsd = None

    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    def _init(self):
        # type: () -> None
        self.task.init(self.conf)

    def launch(self):
        # type: () -> None
        """
        Launch a job by initializing job, run task and publish
        :return:
        """

        logging.info('Launching a job')
        #  Using nested try finally to make sure task get closed as soon as possible as well as to guarantee all the
        #  closeable get closed.
        try:
            is_success = True
            self._init()
            try:
                self.task.run()
            finally:
                self.task.close()

            self.publisher.init(Scoped.get_scoped_conf(self.conf, self.publisher.get_scope()))
            Job.closer.register(self.publisher.close)
            self.publisher.publish()
        except Exception as e:
            is_success = False
            raise e
        finally:
            # TODO: If more metrics are needed on different construct, such as task, consider abstracting this out
            if self.statsd:
                if is_success:
                    LOGGER.info('Publishing job metrics for success')
                    self.statsd.incr('success')
                else:
                    LOGGER.info('Publishing job metrics for failure')
                    self.statsd.incr('fail')

            Job.closer.close()

        logging.info('Job completed')



from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, NODE_KEY, \
    NODE_LABEL, RELATION_START_KEY, RELATION_START_LABEL, RELATION_END_KEY, \
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE

from databuilder.models.table_metadata import TableMetadata


class TableLastUpdated(Neo4jCsvSerializable):
    # constants
    LAST_UPDATED_NODE_LABEL = 'Timestamp'
    LAST_UPDATED_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/timestamp'
    TIMESTAMP_PROPERTY = 'last_updated_timestamp'
    TIMESTAMP_NAME_PROPERTY = 'name'

    TABLE_LASTUPDATED_RELATION_TYPE = 'LAST_UPDATED_AT'
    LASTUPDATED_TABLE_RELATION_TYPE = 'LAST_UPDATED_TIME_OF'

    def __init__(self,
                 table_name,  # type: str
                 last_updated_time_epoch,  # type: int
                 schema_name,  # type: str
                 db='hive',  # type: str
                 cluster='gold'  # type: str
                 ):
        # type: (...) -> None
        self.table_name = table_name
        self.last_updated_time = int(last_updated_time_epoch)
        self.schema = schema_name
        self.db = db
        self.cluster = cluster

        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def __repr__(self):
        # type: (...) -> str
        return \
            """TableLastUpdated(table_name={!r}, last_updated_time={!r}, schema_name={!r}, db={!r}, cluster={!r})"""\
            .format(self.table_name, self.last_updated_time, self.schema, self.db, self.cluster)

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # creates new node
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: (...) -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def get_table_model_key(self):
        # type: (...) -> str
        # returns formatted string for table name
        return TableMetadata.TABLE_KEY_FORMAT.format(db=self.db,
                                                     cluster=self.cluster,
                                                     schema=self.schema,
                                                     tbl=self.table_name)

    def get_last_updated_model_key(self):
        # type: (...) -> str
        # returns formatted string for last updated name
        return TableLastUpdated.LAST_UPDATED_KEY_FORMAT.format(db=self.db,
                                                               cluster=self.cluster,
                                                               schema=self.schema,
                                                               tbl=self.table_name)

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records
        :return:
        """
        results = []

        results.append({
            NODE_KEY: self.get_last_updated_model_key(),
            NODE_LABEL: TableLastUpdated.LAST_UPDATED_NODE_LABEL,
            TableLastUpdated.TIMESTAMP_PROPERTY: self.last_updated_time,
            TableLastUpdated.TIMESTAMP_NAME_PROPERTY: TableLastUpdated.TIMESTAMP_PROPERTY
        })

        return results

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of relations mapping last updated node with table node
        :return:
        """
        results = [{
            RELATION_START_KEY: self.get_table_model_key(),
            RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
            RELATION_END_KEY: self.get_last_updated_model_key(),
            RELATION_END_LABEL: TableLastUpdated.LAST_UPDATED_NODE_LABEL,
            RELATION_TYPE: TableLastUpdated.TABLE_LASTUPDATED_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableLastUpdated.LASTUPDATED_TABLE_RELATION_TYPE
        }]

        return results



from typing import Iterable, Union, Dict, Any, Iterator  # noqa: F401

from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, RELATION_START_KEY, RELATION_END_KEY,
    RELATION_START_LABEL, RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE
)
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.user import User
from databuilder.publisher.neo4j_csv_publisher import UNQUOTED_SUFFIX


class ColumnReader(object):
    """
    A class represent user's read action on column. Implicitly assumes that read count is one.
    """
    def __init__(self,
                 database,  # type: str
                 cluster,  # type: str
                 schema,  # type: str
                 table,  # type: str
                 column,  # type: str
                 user_email,  # type: str
                 read_count=1  # type: int
                 ):
        # type: (...) -> None
        self.database = database.lower()
        self.cluster = cluster.lower()
        self.schema = schema.lower()
        self.table = table.lower()
        self.column = column.lower()
        self.user_email = user_email.lower()
        self.read_count = read_count

    def __repr__(self):
        # type: () -> str
        return """\
ColumnReader(database={!r}, cluster={!r}, schema={!r}, table={!r}, column={!r}, user_email={!r}, read_count={!r})"""\
            .format(self.database, self.cluster, self.schema, self.table, self.column, self.user_email, self.read_count)


class TableColumnUsage(Neo4jCsvSerializable):
    """
    A model represents user <--> column graph model
    Currently it only support to serialize to table level
    """
    TABLE_NODE_LABEL = TableMetadata.TABLE_NODE_LABEL
    TABLE_NODE_KEY_FORMAT = TableMetadata.TABLE_KEY_FORMAT

    USER_TABLE_RELATION_TYPE = 'READ'
    TABLE_USER_RELATION_TYPE = 'READ_BY'

    # Property key for relationship read, readby relationship
    READ_RELATION_COUNT = 'read_count{}'.format(UNQUOTED_SUFFIX)

    def __init__(self,
                 col_readers,  # type: Iterable[ColumnReader]
                 ):
        # type: (...) -> None
        for col_reader in col_readers:
            if col_reader.column != '*':
                raise NotImplementedError('Column is not supported yet {}'.format(col_readers))

        self.col_readers = col_readers
        self._node_iterator = self._create_node_iterator()
        self._rel_iter = self._create_rel_iterator()

    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]

        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self):
        # type: () -> Iterator[Any]
        for col_reader in self.col_readers:
            if col_reader.column == '*':
                # using yield for better memory efficiency
                yield User(email=col_reader.user_email).create_nodes()[0]

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]

        try:
            return next(self._rel_iter)
        except StopIteration:
            return None

    def _create_rel_iterator(self):
        # type: () -> Iterator[Any]
        for col_reader in self.col_readers:
            yield {
                RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
                RELATION_END_LABEL: User.USER_NODE_LABEL,
                RELATION_START_KEY: self._get_table_key(col_reader),
                RELATION_END_KEY: self._get_user_key(col_reader.user_email),
                RELATION_TYPE: TableColumnUsage.TABLE_USER_RELATION_TYPE,
                RELATION_REVERSE_TYPE: TableColumnUsage.USER_TABLE_RELATION_TYPE,
                TableColumnUsage.READ_RELATION_COUNT: col_reader.read_count
            }

    def _get_table_key(self, col_reader):
        # type: (ColumnReader) -> str
        return TableMetadata.TABLE_KEY_FORMAT.format(db=col_reader.database,
                                                     cluster=col_reader.cluster,
                                                     schema=col_reader.schema,
                                                     tbl=col_reader.table)

    def _get_user_key(self, email):
        # type: (str) -> str
        return User.get_user_model_key(email=email)

    def __repr__(self):
        # type: () -> str
        return 'TableColumnUsage(col_readers={!r})'.format(self.col_readers)



from collections import namedtuple

from typing import Iterable, Any, Union, Iterator, Dict, Set  # noqa: F401

from databuilder.models.neo4j_csv_serde import (
    Neo4jCsvSerializable, NODE_LABEL, NODE_KEY, RELATION_START_KEY, RELATION_END_KEY, RELATION_START_LABEL,
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE)
from databuilder.publisher.neo4j_csv_publisher import UNQUOTED_SUFFIX

DESCRIPTION_NODE_LABEL = 'Description'


class ColumnMetadata:
    COLUMN_NODE_LABEL = 'Column'
    COLUMN_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}'
    COLUMN_NAME = 'name'
    COLUMN_TYPE = 'type'
    COLUMN_ORDER = 'sort_order'
    COLUMN_DESCRIPTION = 'description'
    COLUMN_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}/_description'

    # pair of nodes makes relationship where name of variable represents order of relationship.
    COL_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    DESCRIPTION_COL_RELATION_TYPE = 'DESCRIPTION_OF'

    def __init__(self,
                 name,  # type: str
                 description,  # type: Union[str, None]
                 col_type,  # type: str
                 sort_order,  # type: int
                 ):
        # type: (...) -> None
        """
        TODO: Add stats
        :param name:
        :param description:
        :param col_type:
        :param sort_order:
        """
        self.name = name
        self.description = description
        self.type = col_type
        self.sort_order = sort_order

    def __repr__(self):
        # type: () -> str
        return 'ColumnMetadata({!r}, {!r}, {!r}, {!r})'.format(self.name,
                                                               self.description,
                                                               self.type,
                                                               self.sort_order)


# Tuples for de-dupe purpose on Database, Cluster, Schema. See TableMetadata docstring for more information
NodeTuple = namedtuple('KeyName', ['key', 'name', 'label'])
RelTuple = namedtuple('RelKeys', ['start_label', 'end_label', 'start_key', 'end_key', 'type', 'reverse_type'])


class TableMetadata(Neo4jCsvSerializable):
    """
    Table metadata that contains columns. It implements Neo4jCsvSerializable so that it can be serialized to produce
    Table, Column and relation of those along with relationship with table and schema. Additionally, it will create
    Database, Cluster, and Schema with relastionships between those.
    These are being created here as it does not make much sense to have different extraction to produce this. As
    database, cluster, schema would be very repititive with low cardinality, it will perform de-dupe so that publisher
    won't need to publish same nodes, relationships.

    This class can be used for both table and view metadata. If it is a View, is_view=True should be passed in.
    """
    TABLE_NODE_LABEL = 'Table'
    TABLE_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}'
    TABLE_NAME = 'name'
    IS_VIEW = 'is_view{}'.format(UNQUOTED_SUFFIX)  # bool value needs to be unquoted when publish to neo4j

    TABLE_DESCRIPTION = 'description'
    TABLE_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/_description'
    TABLE_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    DESCRIPTION_TABLE_RELATION_TYPE = 'DESCRIPTION_OF'

    DATABASE_NODE_LABEL = 'Database'
    DATABASE_KEY_FORMAT = 'database://{db}'
    DATABASE_CLUSTER_RELATION_TYPE = 'CLUSTER'
    CLUSTER_DATABASE_RELATION_TYPE = 'CLUSTER_OF'

    CLUSTER_NODE_LABEL = 'Cluster'
    CLUSTER_KEY_FORMAT = '{db}://{cluster}'
    CLUSTER_SCHEMA_RELATION_TYPE = 'SCHEMA'
    SCHEMA_CLUSTER_RELATION_TYPE = 'SCHEMA_OF'

    SCHEMA_NODE_LABEL = 'Schema'
    SCHEMA_KEY_FORMAT = '{db}://{cluster}.{schema}'
    SCHEMA_TABLE_RELATION_TYPE = 'TABLE'
    TABLE_SCHEMA_RELATION_TYPE = 'TABLE_OF'

    TABLE_COL_RELATION_TYPE = 'COLUMN'
    COL_TABLE_RELATION_TYPE = 'COLUMN_OF'

    # Only for deduping database, cluster, and schema (table and column will be always processed)
    serialized_nodes = set()  # type: Set[Any]
    serialized_rels = set()  # type: Set[Any]

    def __init__(self,
                 database,  # type: str
                 cluster,  # type: str
                 schema_name,  # type: str
                 name,  # type: str
                 description,  # type: Union[str, None]
                 columns=None,  # type: Iterable[ColumnMetadata]
                 is_view=False,  # type: bool
                 ):
        # type: (...) -> None
        """
        TODO: Add owners
        :param database:
        :param cluster:
        :param schema_name:
        :param name:
        :param description:
        :param columns:
        """
        self.database = database
        self.cluster = cluster
        self.schema_name = schema_name
        self.name = name
        self.description = description
        self.columns = columns if columns else []
        self.is_view = is_view
        self._node_iterator = self._create_next_node()
        self._relation_iterator = self._create_next_relation()

    def __repr__(self):
        # type: () -> str
        return 'TableMetadata({!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r})'.format(self.database,
                                                                                self.cluster,
                                                                                self.schema_name,
                                                                                self.name,
                                                                                self.description,
                                                                                self.columns,
                                                                                self.is_view)

    def _get_table_key(self):
        # type: () -> str
        return TableMetadata.TABLE_KEY_FORMAT.format(db=self.database,
                                                     cluster=self.cluster,
                                                     schema=self.schema_name,
                                                     tbl=self.name)

    def _get_table_description_key(self):
        # type: () -> str
        return TableMetadata.TABLE_DESCRIPTION_FORMAT.format(db=self.database,
                                                             cluster=self.cluster,
                                                             schema=self.schema_name,
                                                             tbl=self.name)

    def _get_database_key(self):
        # type: () -> str
        return TableMetadata.DATABASE_KEY_FORMAT.format(db=self.database)

    def _get_cluster_key(self):
        # type: () -> str
        return TableMetadata.CLUSTER_KEY_FORMAT.format(db=self.database,
                                                       cluster=self.cluster)

    def _get_schema_key(self):
        # type: () -> str
        return TableMetadata.SCHEMA_KEY_FORMAT.format(db=self.database,
                                                      cluster=self.cluster,
                                                      schema=self.schema_name)

    def _get_col_key(self, col):
        # type: (ColumnMetadata) -> str
        return ColumnMetadata.COLUMN_KEY_FORMAT.format(db=self.database,
                                                       cluster=self.cluster,
                                                       schema=self.schema_name,
                                                       tbl=self.name,
                                                       col=col.name)

    def _get_col_description_key(self, col):
        # type: (ColumnMetadata) -> str
        return ColumnMetadata.COLUMN_DESCRIPTION_FORMAT.format(db=self.database,
                                                               cluster=self.cluster,
                                                               schema=self.schema_name,
                                                               tbl=self.name,
                                                               col=col.name)

    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_next_node(self):
        # type: () -> Iterator[Any]
        yield {NODE_LABEL: TableMetadata.TABLE_NODE_LABEL,
               NODE_KEY: self._get_table_key(),
               TableMetadata.TABLE_NAME: self.name,
               TableMetadata.IS_VIEW: self.is_view}

        if self.description:
            yield {NODE_LABEL: DESCRIPTION_NODE_LABEL,
                   NODE_KEY: self._get_table_description_key(),
                   TableMetadata.TABLE_DESCRIPTION: self.description}

        for col in self.columns:
            yield {
                NODE_LABEL: ColumnMetadata.COLUMN_NODE_LABEL,
                NODE_KEY: self._get_col_key(col),
                ColumnMetadata.COLUMN_NAME: col.name,
                ColumnMetadata.COLUMN_TYPE: col.type,
                ColumnMetadata.COLUMN_ORDER: col.sort_order}

            if not col.description:
                continue

            yield {
                NODE_LABEL: DESCRIPTION_NODE_LABEL,
                NODE_KEY: self._get_col_description_key(col),
                ColumnMetadata.COLUMN_DESCRIPTION: col.description}

        # Database, cluster, schema
        others = [NodeTuple(key=self._get_database_key(),
                            name=self.database,
                            label=TableMetadata.DATABASE_NODE_LABEL),
                  NodeTuple(key=self._get_cluster_key(),
                            name=self.cluster,
                            label=TableMetadata.CLUSTER_NODE_LABEL),
                  NodeTuple(key=self._get_schema_key(),
                            name=self.schema_name,
                            label=TableMetadata.SCHEMA_NODE_LABEL)
                  ]

        for node_tuple in others:
            if node_tuple not in TableMetadata.serialized_nodes:
                TableMetadata.serialized_nodes.add(node_tuple)
                yield {
                    NODE_LABEL: node_tuple.label,
                    NODE_KEY: node_tuple.key,
                    'name': node_tuple.name
                }

    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_next_relation(self):
        # type: () -> Iterator[Any]

        yield {
            RELATION_START_LABEL: TableMetadata.SCHEMA_NODE_LABEL,
            RELATION_END_LABEL: TableMetadata.TABLE_NODE_LABEL,
            RELATION_START_KEY: self._get_schema_key(),
            RELATION_END_KEY: self._get_table_key(),
            RELATION_TYPE: TableMetadata.SCHEMA_TABLE_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableMetadata.TABLE_SCHEMA_RELATION_TYPE
        }

        if self.description:
            yield {
                RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
                RELATION_END_LABEL: DESCRIPTION_NODE_LABEL,
                RELATION_START_KEY: self._get_table_key(),
                RELATION_END_KEY: self._get_table_description_key(),
                RELATION_TYPE: TableMetadata.TABLE_DESCRIPTION_RELATION_TYPE,
                RELATION_REVERSE_TYPE: TableMetadata.DESCRIPTION_TABLE_RELATION_TYPE
            }

        for col in self.columns:
            yield {
                RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
                RELATION_END_LABEL: ColumnMetadata.COLUMN_NODE_LABEL,
                RELATION_START_KEY: self._get_table_key(),
                RELATION_END_KEY: self._get_col_key(col),
                RELATION_TYPE: TableMetadata.TABLE_COL_RELATION_TYPE,
                RELATION_REVERSE_TYPE: TableMetadata.COL_TABLE_RELATION_TYPE
            }

            if not col.description:
                continue

            yield {
                RELATION_START_LABEL: ColumnMetadata.COLUMN_NODE_LABEL,
                RELATION_END_LABEL: DESCRIPTION_NODE_LABEL,
                RELATION_START_KEY: self._get_col_key(col),
                RELATION_END_KEY: self._get_col_description_key(col),
                RELATION_TYPE: ColumnMetadata.COL_DESCRIPTION_RELATION_TYPE,
                RELATION_REVERSE_TYPE: ColumnMetadata.DESCRIPTION_COL_RELATION_TYPE
            }

        others = [
            RelTuple(start_label=TableMetadata.DATABASE_NODE_LABEL,
                     end_label=TableMetadata.CLUSTER_NODE_LABEL,
                     start_key=self._get_database_key(),
                     end_key=self._get_cluster_key(),
                     type=TableMetadata.DATABASE_CLUSTER_RELATION_TYPE,
                     reverse_type=TableMetadata.CLUSTER_DATABASE_RELATION_TYPE),
            RelTuple(start_label=TableMetadata.CLUSTER_NODE_LABEL,
                     end_label=TableMetadata.SCHEMA_NODE_LABEL,
                     start_key=self._get_cluster_key(),
                     end_key=self._get_schema_key(),
                     type=TableMetadata.CLUSTER_SCHEMA_RELATION_TYPE,
                     reverse_type=TableMetadata.SCHEMA_CLUSTER_RELATION_TYPE)
        ]

        for rel_tuple in others:
            if rel_tuple not in TableMetadata.serialized_rels:
                TableMetadata.serialized_rels.add(rel_tuple)
                yield {
                    RELATION_START_LABEL: rel_tuple.start_label,
                    RELATION_END_LABEL: rel_tuple.end_label,
                    RELATION_START_KEY: rel_tuple.start_key,
                    RELATION_END_KEY: rel_tuple.end_key,
                    RELATION_TYPE: rel_tuple.type,
                    RELATION_REVERSE_TYPE: rel_tuple.reverse_type
                }



import copy
import csv
import logging
import time
from os import listdir
from os.path import isfile, join
from string import Template

import six
from neo4j.v1 import GraphDatabase, Transaction  # noqa: F401
from pyhocon import ConfigFactory  # noqa: F401
from pyhocon import ConfigTree  # noqa: F401
from typing import Set, List  # noqa: F401

from databuilder.publisher.base_publisher import Publisher

# Config keys
# A directory that contains CSV files for nodes
NODE_FILES_DIR = 'node_files_directory'
# A directory that contains CSV files for relationships
RELATION_FILES_DIR = 'relation_files_directory'
# A end point for Neo4j e.g: bolt://localhost:9999
NEO4J_END_POINT_KEY = 'neo4j_endpoint'
# A transaction size that determines how often it commits.
NEO4J_TRANSCATION_SIZE = 'neo4j_transaction_size'
# A boolean flag to make it fail if relationship is not created
NEO4J_RELATIONSHIP_CREATION_CONFIRM = 'neo4j_relationship_creation_confirm'

NEO4J_MAX_CONN_LIFE_TIME_SEC = 'neo4j_max_conn_life_time_sec'

# list of nodes that are create only, and not updated if match exists
NEO4J_CREATE_ONLY_NODES = 'neo4j_create_only_nodes'

NEO4J_USER = 'neo4j_user'
NEO4J_PASSWORD = 'neo4j_password'

# This will be used to provide unique tag to the node and relationship
JOB_PUBLISH_TAG = 'job_publish_tag'

# Neo4j property name for published tag
PUBLISHED_TAG_PROPERTY_NAME = 'published_tag'

# CSV HEADER
# A header with this suffix will be pass to Neo4j statement without quote
UNQUOTED_SUFFIX = ':UNQUOTED'
# A header for Node label
NODE_LABEL_KEY = 'LABEL'
# A header for Node key
NODE_KEY_KEY = 'KEY'
# Required columns for Node
NODE_REQUIRED_KEYS = {NODE_LABEL_KEY, NODE_KEY_KEY}

# Relationship relates two nodes together
# Start node label
RELATION_START_LABEL = 'START_LABEL'
# Start node key
RELATION_START_KEY = 'START_KEY'
# End node label
RELATION_END_LABEL = 'END_LABEL'
# Node node key
RELATION_END_KEY = 'END_KEY'
# Type for relationship (Start Node)->(End Node)
RELATION_TYPE = 'TYPE'
# Type for reverse relationship (End Node)->(Start Node)
RELATION_REVERSE_TYPE = 'REVERSE_TYPE'
# Required columns for Relationship
RELATION_REQUIRED_KEYS = {RELATION_START_LABEL, RELATION_START_KEY,
                          RELATION_END_LABEL, RELATION_END_KEY,
                          RELATION_TYPE, RELATION_REVERSE_TYPE}


DEFAULT_CONFIG = ConfigFactory.from_dict({NEO4J_TRANSCATION_SIZE: 500,
                                          NEO4J_RELATIONSHIP_CREATION_CONFIRM: False,
                                          NEO4J_MAX_CONN_LIFE_TIME_SEC: 50})

NODE_MERGE_TEMPLATE = Template("""MERGE (node:$LABEL {key: '${KEY}'})
ON CREATE SET ${create_prop_body}
${update_statement}""")

NODE_UPDATE_TEMPLATE = Template("""ON MATCH SET ${update_prop_body}""")

RELATION_MERGE_TEMPLATE = Template("""MATCH (n1:$START_LABEL {key: '${START_KEY}'}),
(n2:$END_LABEL {key: '${END_KEY}'})
MERGE (n1)-[r1:$TYPE]->(n2)-[r2:$REVERSE_TYPE]->(n1)
$PROP_STMT RETURN n1.key, n2.key""")

CREATE_UNIQUE_INDEX_TEMPLATE = Template('CREATE CONSTRAINT ON (node:${LABEL}) ASSERT node.key IS UNIQUE')

LOGGER = logging.getLogger(__name__)


class Neo4jCsvPublisher(Publisher):
    """
    A Publisher takes two folders for input and publishes to Neo4j.
    One folder will contain CSV file(s) for Node where the other folder will contain CSV file(s) for Relationship.

    Neo4j follows Label Node properties Graph and more information about this is in:
    https://neo4j.com/docs/developer-manual/current/introduction/graphdb-concepts/

    #TODO User UNWIND batch operation for better performance
    """
    def __init__(self):
        # type: () -> None
        pass

    def init(self, conf):
        # type: (ConfigTree) -> None
        conf = conf.with_fallback(DEFAULT_CONFIG)

        self._node_files = self._list_files(conf, NODE_FILES_DIR)
        self._node_files_iter = iter(self._node_files)

        self._relation_files = self._list_files(conf, RELATION_FILES_DIR)
        self._relation_files_iter = iter(self._relation_files)

        self._driver = \
            GraphDatabase.driver(conf.get_string(NEO4J_END_POINT_KEY),
                                 max_connection_life_time=50,
                                 auth=(conf.get_string(NEO4J_USER), conf.get_string(NEO4J_PASSWORD)))
        self._transaction_size = conf.get_int(NEO4J_TRANSCATION_SIZE)
        self._session = self._driver.session()
        self._confirm_rel_created = conf.get_bool(NEO4J_RELATIONSHIP_CREATION_CONFIRM)

        # config is list of node label.
        # When set, this list specifies a list of nodes that shouldn't be updated, if exists
        self.create_only_nodes = set(conf.get_list(NEO4J_CREATE_ONLY_NODES, default=[]))
        self.labels = set()  # type: Set[str]
        self.publish_tag = conf.get_string(JOB_PUBLISH_TAG)  # type: str
        if not self.publish_tag:
            raise Exception('{} should not be empty'.format(JOB_PUBLISH_TAG))

        LOGGER.info('Publishing Node csv files {}, and Relation CSV files {}'
                    .format(self._node_files, self._relation_files))

    def _list_files(self, conf, path_key):
        # type: (ConfigTree, str) -> List[str]
        """
        List files from directory
        :param conf:
        :param path_key:
        :return: List of file paths
        """
        if path_key not in conf:
            return []

        path = conf.get_string(path_key)
        return [join(path, f) for f in listdir(path) if isfile(join(path, f))]

    def publish(self):
        # type: () -> None
        """
        Publishes Nodes first and then Relations
        :return:
        """

        start = time.time()

        LOGGER.info('Publishing Node files: {}'.format(self._node_files))
        while True:
            try:
                node_file = next(self._node_files_iter)
                self._publish_node(node_file)
            except StopIteration:
                break

        LOGGER.info('Publishing Relationship files: {}'.format(self._relation_files))
        while True:
            try:
                relation_file = next(self._relation_files_iter)
                self._publish_relation(relation_file)
            except StopIteration:
                break

        # TODO: Add statsd support
        LOGGER.info('Successfully published. Elapsed: {} seconds'.format(time.time() - start))

    def get_scope(self):
        # type: () -> str
        return 'publisher.neo4j'

    def _publish_node(self, node_file):
        # type: (str) -> None
        """
        Iterate over the csv records of a file, each csv record transform to Merge statement and will be executed.
        All nodes should have a unique key, and this method will try to create unique index on the LABEL when it sees
        first time within a job scope.
        Example of Cypher query executed by this method:
        MERGE (col_test_id1:Column {key: 'presto://gold.test_schema1/test_table1/test_id1'})
        ON CREATE SET col_test_id1.name = 'test_id1',
                      col_test_id1.order_pos = 2,
                      col_test_id1.type = 'bigint'
        ON MATCH SET col_test_id1.name = 'test_id1',
                     col_test_id1.order_pos = 2,
                     col_test_id1.type = 'bigint'

        :param node_file:
        :return:
        """
        tx = self._session.begin_transaction()
        with open(node_file, 'r') as node_csv:
            for count, node_record in enumerate(csv.DictReader(node_csv)):
                label = node_record[NODE_LABEL_KEY]
                # If label is seen for the first time, try creating unique index
                if label not in self.labels:
                    tx.commit()  # Transaction needs to be committed as index update will make transaction to abort.
                    tx.close()
                    LOGGER.info('Committed {} records'.format(count + 1))

                    self._try_create_index(label)
                    self.labels.add(label)
                    tx = self._session.begin_transaction()

                stmt = self.create_node_merge_statement(node_record=node_record)
                tx = self._execute_statement(stmt, tx, count)

        tx.commit()
        tx.close()
        LOGGER.info('Committed {} records'.format(count + 1))

    def is_create_only_node(self, node_record):
        # type: (dict) -> bool
        """
        Check if node can be updated
        :param node_record:
        :return:
        """
        if self.create_only_nodes:
            return node_record[NODE_LABEL_KEY] in self.create_only_nodes
        else:
            return False

    def create_node_merge_statement(self, node_record):
        # type: (dict) -> str
        """
        Creates node merge statement
        :param node_record:
        :return:
        """
        params = copy.deepcopy(node_record)
        params['create_prop_body'] = self._create_props_body(node_record, NODE_REQUIRED_KEYS, 'node')

        update_statement = ''
        if not self.is_create_only_node(node_record):
            update_prop_body = self._create_props_body(node_record, NODE_REQUIRED_KEYS, 'node')
            update_statement = NODE_UPDATE_TEMPLATE.substitute(update_prop_body=update_prop_body)
        params['update_statement'] = update_statement

        return NODE_MERGE_TEMPLATE.substitute(params)

    def _publish_relation(self, relation_file):
        # type: (str) -> None
        """
        Creates relation between two nodes.
        (In Amundsen, all relation is bi-directional)

        Example of Cypher query executed by this method:
        MATCH (n1:Table {key: 'presto://gold.test_schema1/test_table1'}),
              (n2:Column {key: 'presto://gold.test_schema1/test_table1/test_col1'})
        MERGE (n1)-[r1:COLUMN]->(n2)-[r2:BELONG_TO_TABLE]->(n1)
        RETURN n1.key, n2.key

        :param relation_file:
        :return:
        """

        tx = self._session.begin_transaction()
        with open(relation_file, 'r') as relation_csv:
            for count, rel_record in enumerate(csv.DictReader(relation_csv)):
                stmt = self.create_relationship_merge_statement(rel_record=rel_record)
                tx = self._execute_statement(stmt, tx, count,
                                             expect_result=self._confirm_rel_created)

        tx.commit()
        tx.close()
        LOGGER.info('Committed {} records'.format(count + 1))

    def create_relationship_merge_statement(self, rel_record):
        # type: (dict) -> str
        """
        Creates relationship merge statement
        :param rel_record:
        :return:
        """
        param = copy.deepcopy(rel_record)
        create_prop_body = self._create_props_body(rel_record, RELATION_REQUIRED_KEYS, 'r1')
        param['PROP_STMT'] = ' '  # No properties for relationship by default

        if create_prop_body:
            # We need one more body for reverse relation
            create_prop_body = ' , '.join([create_prop_body,
                                           self._create_props_body(rel_record, RELATION_REQUIRED_KEYS, 'r2')])
            update_prop_body = ' , '.join([self._create_props_body(rel_record, RELATION_REQUIRED_KEYS, 'r1'),
                                           self._create_props_body(rel_record, RELATION_REQUIRED_KEYS, 'r2')])

            param['PROP_STMT'] = """ON CREATE SET {create_prop_body}
ON MATCH SET {update_prop_body}""".format(create_prop_body=create_prop_body,
                                          update_prop_body=update_prop_body)

        return RELATION_MERGE_TEMPLATE.substitute(param)

    def _create_props_body(self,
                           record_dict,
                           excludes,
                           identifier):
        # type: (dict, Set, str) -> str
        """
        Creates properties body with params required for resolving template.

        e.g: Note that node.key3 is not quoted if header has UNQUOTED_SUFFIX.
        identifier.key1 = 'val1' , identifier.key2 = 'val2', identifier.key3 = val3

        :param record_dict: A dict represents CSV row
        :param excludes: set of excluded columns that does not need to be in properties (e.g: KEY, LABEL ...)
        :param identifier: identifier that will be used in CYPHER query as shown on above example
        :param is_update: Creates property body for update statement in MERGE
        :return: Properties body for Cypher statement
        """
        template_params = {}
        props = []
        for k, v in six.iteritems(record_dict):
            if k in excludes:
                template_params[k] = v
                continue

            # escape quote for Cypher query
            # if isinstance(str, v):
            v = v.replace('\'', "\\'")

            if k.endswith(UNQUOTED_SUFFIX):
                k = k[:-len(UNQUOTED_SUFFIX)]
                props.append('{id}.{key} = {val}'.format(id=identifier, key=k, val=v))
            else:
                props.append("""{id}.{key} = '{val}'""".format(id=identifier, key=k, val=v))

            template_params[k] = v

        props.append("""{id}.{key} = '{val}'""".format(id=identifier,
                                                       key=PUBLISHED_TAG_PROPERTY_NAME,
                                                       val=self.publish_tag))

        return ', '.join(props)

    def _execute_statement(self,
                           stmt,
                           tx,
                           count,
                           expect_result=False):
        # type: (str, Transaction, int, bool) -> Transaction

        """
        Executes statement against Neo4j. If execution fails, it rollsback and raise exception.
        If 'expect_result' flag is True, it confirms if result object is not null.
        :param stmt:
        :param tx:
        :param count:
        :param expect_result: By having this True, it will validate if result object is not None.
        :return:
        """
        try:
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug('Executing statement: {}'.format(stmt))

            if six.PY2:
                result = tx.run(unicode(stmt, errors='ignore')) # noqa
            else:
                result = tx.run(str(stmt).encode('utf-8', 'ignore'))
            if expect_result and not result.single():
                raise RuntimeError('Failed to executed statement: {}'.format(stmt))

            if count > 1 and count % self._transaction_size == 0:
                tx.commit()
                tx.close()
                LOGGER.info('Committed {} records so far'.format(count))
                return self._session.begin_transaction()

            return tx
        except Exception as e:
            LOGGER.exception('Failed to execute Cypher query')
            if not tx.closed():
                tx.rollback()
            raise e

    def _try_create_index(self,
                          label):
        # type: (str) -> None
        """
        For any label seen first time for this publisher it will try to create unique index.
        There's no side effect on Neo4j side issuing index creation for existing index.
        :param label:
        :return:
        """
        stmt = CREATE_UNIQUE_INDEX_TEMPLATE.substitute(LABEL=label)
        LOGGER.info('Trying to create index for label {label} if not exist: {stmt}'.format(label=label,
                                                                                           stmt=stmt))
        with self._driver.session() as session:
            session.run(stmt)



from collections import namedtuple
import logging

from pyhocon import ConfigTree  # noqa: F401
from typing import Dict, List, Any, Optional  # noqa: F401

from databuilder import Scoped
from databuilder.transformer.base_transformer import NoopTransformer
from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_column_usage import ColumnReader, TableColumnUsage
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.regex_str_replace_transformer import RegexStrReplaceTransformer
from databuilder.transformer.sql_to_table_col_usage_transformer import SqlToTblColUsageTransformer

TableColumnUsageTuple = namedtuple('TableColumnUsageTuple', ['database', 'cluster', 'schema',
                                                             'table', 'column', 'email'])

LOGGER = logging.getLogger(__name__)


# Config keys:
RAW_EXTRACTOR = 'raw_extractor'


class TblColUsgAggExtractor(Extractor):
    """
    An aggregate extractor for table column usage.
    It uses RegexStrReplaceTransformer to cleanse SQL statement and uses SqlToTblColUsageTransformer to get table
    column usage.

    All usage will be aggregated in memory and on last record, it will return aggregated TableColumnUsage
    Note that this extractor will do all the transformation and aggregation so that no more transformation is needed,
    after this.
    """

    def init(self, conf):
        # type: (ConfigTree) -> None
        self._extractor = conf.get(RAW_EXTRACTOR)  # type: Extractor
        self._extractor.init(Scoped.get_scoped_conf(conf, self._extractor.get_scope()))

        regex_transformer = RegexStrReplaceTransformer()  # type: Any
        if conf.get(regex_transformer.get_scope(), None):
            regex_transformer.init(Scoped.get_scoped_conf(conf, regex_transformer.get_scope()))
        else:
            LOGGER.info('{} is not defined. Not using it'.format(regex_transformer.get_scope()))
            regex_transformer = NoopTransformer()

        sql_to_usage_transformer = SqlToTblColUsageTransformer()
        sql_to_usage_transformer.init(Scoped.get_scoped_conf(conf, sql_to_usage_transformer.get_scope()))

        self._transformer = ChainedTransformer((regex_transformer, sql_to_usage_transformer))

    def extract(self):
        # type: () -> Optional[TableColumnUsage]
        """
        It aggregates all count per table and user in memory. Table level aggregation don't expect to occupy much
        memory.
        :return: Provides a record or None if no more to extract
        """
        count_map = {}  # type: Dict[TableColumnUsageTuple, int]
        record = self._extractor.extract()

        count = 0
        while record:
            count += 1
            if count % 1000 == 0:
                LOGGER.info('Aggregated {} records'.format(count))

            tbl_col_usg = self._transformer.transform(record=record)
            record = self._extractor.extract()
            # filtered case
            if not tbl_col_usg:
                continue

            for col_rdr in tbl_col_usg.col_readers:
                key = TableColumnUsageTuple(database=col_rdr.database, cluster=col_rdr.cluster, schema=col_rdr.schema,
                                            table=col_rdr.table, column=col_rdr.column, email=col_rdr.user_email)
                new_count = count_map.get(key, 0) + col_rdr.read_count
                count_map[key] = new_count

        if not len(count_map):
            return None

        col_readers = []  # type: List[ColumnReader]

        while len(count_map):
            tbl_col_rdr_tuple, count = count_map.popitem()
            col_readers.append(ColumnReader(database=tbl_col_rdr_tuple.database, cluster=tbl_col_rdr_tuple.cluster,
                                            schema=tbl_col_rdr_tuple.schema, table=tbl_col_rdr_tuple.table,
                                            column=tbl_col_rdr_tuple.column, user_email=tbl_col_rdr_tuple.email,
                                            read_count=count))

        return TableColumnUsage(col_readers=col_readers)

    def get_scope(self):
        # type: () -> str
        return 'extractor.table_column_usage_aggregate'

    def close(self):
        # type: () -> None
        self._transformer.close()



import abc

from pyhocon import ConfigTree  # noqa: F401

from databuilder import Scoped


class Publisher(Scoped):
    """
    A Publisher that writes dataset (not a record) in Atomic manner,
    if possible.
    (Either success or fail, no partial state)
    Use case: If you want to use neo4j import util or Load CSV util,
    that takes CSV file to load database, you need to first create CSV file.
    CSV file holds number of records, and loader can writes multiple records
    to it. Once loader finishes writing CSV file, you have complete CSV file,
    ready to publish to Neo4j. Publisher can take the location of CSV file,
    and push to Neo4j.

    """
    @abc.abstractmethod
    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    @abc.abstractmethod
    def publish(self):
        # type: () -> None
        pass

    def get_scope(self):
        # type: () -> str
        return 'publisher'


class NoopPublisher(Publisher):
    def __init__(self):
        # type: () -> None
        pass

    def init(self, conf):
        # type: (ConfigTree) -> None
        pass

    def publish(self):
        # type: () -> None
        pass

    def get_scope(self):
        # type: () -> str
        return 'publisher.noop'



# Generated from SqlBase.g4 by ANTLR 4.7.1
from antlr4 import *

# This class defines a complete listener for a parse tree produced by SqlBaseParser.
class SqlBaseListener(ParseTreeListener):

    # Enter a parse tree produced by SqlBaseParser#singleStatement.
    def enterSingleStatement(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#singleStatement.
    def exitSingleStatement(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#singleExpression.
    def enterSingleExpression(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#singleExpression.
    def exitSingleExpression(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#statementDefault.
    def enterStatementDefault(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#statementDefault.
    def exitStatementDefault(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#use.
    def enterUse(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#use.
    def exitUse(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#createSchema.
    def enterCreateSchema(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#createSchema.
    def exitCreateSchema(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#dropSchema.
    def enterDropSchema(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#dropSchema.
    def exitDropSchema(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#renameSchema.
    def enterRenameSchema(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#renameSchema.
    def exitRenameSchema(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#createTableAsSelect.
    def enterCreateTableAsSelect(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#createTableAsSelect.
    def exitCreateTableAsSelect(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#createTable.
    def enterCreateTable(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#createTable.
    def exitCreateTable(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#dropTable.
    def enterDropTable(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#dropTable.
    def exitDropTable(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#insertInto.
    def enterInsertInto(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#insertInto.
    def exitInsertInto(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#delete.
    def enterDelete(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#delete.
    def exitDelete(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#renameTable.
    def enterRenameTable(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#renameTable.
    def exitRenameTable(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#renameColumn.
    def enterRenameColumn(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#renameColumn.
    def exitRenameColumn(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#dropColumn.
    def enterDropColumn(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#dropColumn.
    def exitDropColumn(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#addColumn.
    def enterAddColumn(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#addColumn.
    def exitAddColumn(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#createView.
    def enterCreateView(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#createView.
    def exitCreateView(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#dropView.
    def enterDropView(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#dropView.
    def exitDropView(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#call.
    def enterCall(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#call.
    def exitCall(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#grant.
    def enterGrant(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#grant.
    def exitGrant(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#revoke.
    def enterRevoke(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#revoke.
    def exitRevoke(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showGrants.
    def enterShowGrants(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showGrants.
    def exitShowGrants(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#explain.
    def enterExplain(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#explain.
    def exitExplain(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showCreateTable.
    def enterShowCreateTable(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showCreateTable.
    def exitShowCreateTable(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showCreateView.
    def enterShowCreateView(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showCreateView.
    def exitShowCreateView(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showTables.
    def enterShowTables(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showTables.
    def exitShowTables(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showSchemas.
    def enterShowSchemas(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showSchemas.
    def exitShowSchemas(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showCatalogs.
    def enterShowCatalogs(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showCatalogs.
    def exitShowCatalogs(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showColumns.
    def enterShowColumns(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showColumns.
    def exitShowColumns(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showStats.
    def enterShowStats(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showStats.
    def exitShowStats(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showStatsForQuery.
    def enterShowStatsForQuery(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showStatsForQuery.
    def exitShowStatsForQuery(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showFunctions.
    def enterShowFunctions(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showFunctions.
    def exitShowFunctions(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showSession.
    def enterShowSession(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showSession.
    def exitShowSession(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#setSession.
    def enterSetSession(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#setSession.
    def exitSetSession(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#resetSession.
    def enterResetSession(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#resetSession.
    def exitResetSession(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#startTransaction.
    def enterStartTransaction(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#startTransaction.
    def exitStartTransaction(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#commit.
    def enterCommit(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#commit.
    def exitCommit(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#rollback.
    def enterRollback(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#rollback.
    def exitRollback(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#showPartitions.
    def enterShowPartitions(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#showPartitions.
    def exitShowPartitions(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#prepare.
    def enterPrepare(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#prepare.
    def exitPrepare(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#deallocate.
    def enterDeallocate(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#deallocate.
    def exitDeallocate(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#execute.
    def enterExecute(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#execute.
    def exitExecute(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#describeInput.
    def enterDescribeInput(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#describeInput.
    def exitDescribeInput(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#describeOutput.
    def enterDescribeOutput(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#describeOutput.
    def exitDescribeOutput(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#query.
    def enterQuery(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#query.
    def exitQuery(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#with.
    def enterWith(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#with.
    def exitWith(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#tableElement.
    def enterTableElement(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#tableElement.
    def exitTableElement(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#columnDefinition.
    def enterColumnDefinition(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#columnDefinition.
    def exitColumnDefinition(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#likeClause.
    def enterLikeClause(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#likeClause.
    def exitLikeClause(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#properties.
    def enterProperties(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#properties.
    def exitProperties(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#property.
    def enterProperty(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#property.
    def exitProperty(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#queryNoWith.
    def enterQueryNoWith(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#queryNoWith.
    def exitQueryNoWith(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#queryTermDefault.
    def enterQueryTermDefault(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#queryTermDefault.
    def exitQueryTermDefault(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#setOperation.
    def enterSetOperation(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#setOperation.
    def exitSetOperation(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#queryPrimaryDefault.
    def enterQueryPrimaryDefault(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#queryPrimaryDefault.
    def exitQueryPrimaryDefault(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#table.
    def enterTable(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#table.
    def exitTable(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#inlineTable.
    def enterInlineTable(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#inlineTable.
    def exitInlineTable(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#subquery.
    def enterSubquery(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#subquery.
    def exitSubquery(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#sortItem.
    def enterSortItem(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#sortItem.
    def exitSortItem(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#querySpecification.
    def enterQuerySpecification(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#querySpecification.
    def exitQuerySpecification(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#groupBy.
    def enterGroupBy(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#groupBy.
    def exitGroupBy(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#singleGroupingSet.
    def enterSingleGroupingSet(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#singleGroupingSet.
    def exitSingleGroupingSet(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#rollup.
    def enterRollup(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#rollup.
    def exitRollup(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#cube.
    def enterCube(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#cube.
    def exitCube(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#multipleGroupingSets.
    def enterMultipleGroupingSets(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#multipleGroupingSets.
    def exitMultipleGroupingSets(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#groupingExpressions.
    def enterGroupingExpressions(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#groupingExpressions.
    def exitGroupingExpressions(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#groupingSet.
    def enterGroupingSet(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#groupingSet.
    def exitGroupingSet(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#namedQuery.
    def enterNamedQuery(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#namedQuery.
    def exitNamedQuery(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#setQuantifier.
    def enterSetQuantifier(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#setQuantifier.
    def exitSetQuantifier(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#selectSingle.
    def enterSelectSingle(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#selectSingle.
    def exitSelectSingle(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#selectAll.
    def enterSelectAll(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#selectAll.
    def exitSelectAll(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#relationDefault.
    def enterRelationDefault(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#relationDefault.
    def exitRelationDefault(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#joinRelation.
    def enterJoinRelation(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#joinRelation.
    def exitJoinRelation(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#joinType.
    def enterJoinType(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#joinType.
    def exitJoinType(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#joinCriteria.
    def enterJoinCriteria(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#joinCriteria.
    def exitJoinCriteria(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#sampledRelation.
    def enterSampledRelation(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#sampledRelation.
    def exitSampledRelation(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#sampleType.
    def enterSampleType(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#sampleType.
    def exitSampleType(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#aliasedRelation.
    def enterAliasedRelation(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#aliasedRelation.
    def exitAliasedRelation(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#columnAliases.
    def enterColumnAliases(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#columnAliases.
    def exitColumnAliases(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#tableName.
    def enterTableName(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#tableName.
    def exitTableName(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#subqueryRelation.
    def enterSubqueryRelation(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#subqueryRelation.
    def exitSubqueryRelation(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#unnest.
    def enterUnnest(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#unnest.
    def exitUnnest(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#lateral.
    def enterLateral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#lateral.
    def exitLateral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#parenthesizedRelation.
    def enterParenthesizedRelation(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#parenthesizedRelation.
    def exitParenthesizedRelation(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#expression.
    def enterExpression(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#expression.
    def exitExpression(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#logicalNot.
    def enterLogicalNot(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#logicalNot.
    def exitLogicalNot(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#booleanDefault.
    def enterBooleanDefault(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#booleanDefault.
    def exitBooleanDefault(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#logicalBinary.
    def enterLogicalBinary(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#logicalBinary.
    def exitLogicalBinary(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#predicated.
    def enterPredicated(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#predicated.
    def exitPredicated(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#comparison.
    def enterComparison(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#comparison.
    def exitComparison(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#quantifiedComparison.
    def enterQuantifiedComparison(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#quantifiedComparison.
    def exitQuantifiedComparison(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#between.
    def enterBetween(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#between.
    def exitBetween(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#inList.
    def enterInList(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#inList.
    def exitInList(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#inSubquery.
    def enterInSubquery(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#inSubquery.
    def exitInSubquery(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#like.
    def enterLike(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#like.
    def exitLike(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#nullPredicate.
    def enterNullPredicate(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#nullPredicate.
    def exitNullPredicate(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#distinctFrom.
    def enterDistinctFrom(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#distinctFrom.
    def exitDistinctFrom(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#valueExpressionDefault.
    def enterValueExpressionDefault(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#valueExpressionDefault.
    def exitValueExpressionDefault(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#concatenation.
    def enterConcatenation(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#concatenation.
    def exitConcatenation(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#arithmeticBinary.
    def enterArithmeticBinary(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#arithmeticBinary.
    def exitArithmeticBinary(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#arithmeticUnary.
    def enterArithmeticUnary(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#arithmeticUnary.
    def exitArithmeticUnary(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#atTimeZone.
    def enterAtTimeZone(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#atTimeZone.
    def exitAtTimeZone(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#dereference.
    def enterDereference(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#dereference.
    def exitDereference(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#typeConstructor.
    def enterTypeConstructor(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#typeConstructor.
    def exitTypeConstructor(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#specialDateTimeFunction.
    def enterSpecialDateTimeFunction(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#specialDateTimeFunction.
    def exitSpecialDateTimeFunction(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#substring.
    def enterSubstring(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#substring.
    def exitSubstring(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#cast.
    def enterCast(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#cast.
    def exitCast(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#lambda.
    def enterLambda(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#lambda.
    def exitLambda(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#parenthesizedExpression.
    def enterParenthesizedExpression(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#parenthesizedExpression.
    def exitParenthesizedExpression(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#parameter.
    def enterParameter(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#parameter.
    def exitParameter(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#normalize.
    def enterNormalize(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#normalize.
    def exitNormalize(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#intervalLiteral.
    def enterIntervalLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#intervalLiteral.
    def exitIntervalLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#numericLiteral.
    def enterNumericLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#numericLiteral.
    def exitNumericLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#booleanLiteral.
    def enterBooleanLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#booleanLiteral.
    def exitBooleanLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#simpleCase.
    def enterSimpleCase(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#simpleCase.
    def exitSimpleCase(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#columnReference.
    def enterColumnReference(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#columnReference.
    def exitColumnReference(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#nullLiteral.
    def enterNullLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#nullLiteral.
    def exitNullLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#rowConstructor.
    def enterRowConstructor(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#rowConstructor.
    def exitRowConstructor(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#subscript.
    def enterSubscript(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#subscript.
    def exitSubscript(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#subqueryExpression.
    def enterSubqueryExpression(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#subqueryExpression.
    def exitSubqueryExpression(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#binaryLiteral.
    def enterBinaryLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#binaryLiteral.
    def exitBinaryLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#currentUser.
    def enterCurrentUser(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#currentUser.
    def exitCurrentUser(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#extract.
    def enterExtract(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#extract.
    def exitExtract(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#stringLiteral.
    def enterStringLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#stringLiteral.
    def exitStringLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#arrayConstructor.
    def enterArrayConstructor(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#arrayConstructor.
    def exitArrayConstructor(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#functionCall.
    def enterFunctionCall(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#functionCall.
    def exitFunctionCall(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#exists.
    def enterExists(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#exists.
    def exitExists(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#position.
    def enterPosition(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#position.
    def exitPosition(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#searchedCase.
    def enterSearchedCase(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#searchedCase.
    def exitSearchedCase(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#groupingOperation.
    def enterGroupingOperation(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#groupingOperation.
    def exitGroupingOperation(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#basicStringLiteral.
    def enterBasicStringLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#basicStringLiteral.
    def exitBasicStringLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#unicodeStringLiteral.
    def enterUnicodeStringLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#unicodeStringLiteral.
    def exitUnicodeStringLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#timeZoneInterval.
    def enterTimeZoneInterval(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#timeZoneInterval.
    def exitTimeZoneInterval(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#timeZoneString.
    def enterTimeZoneString(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#timeZoneString.
    def exitTimeZoneString(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#comparisonOperator.
    def enterComparisonOperator(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#comparisonOperator.
    def exitComparisonOperator(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#comparisonQuantifier.
    def enterComparisonQuantifier(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#comparisonQuantifier.
    def exitComparisonQuantifier(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#booleanValue.
    def enterBooleanValue(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#booleanValue.
    def exitBooleanValue(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#interval.
    def enterInterval(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#interval.
    def exitInterval(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#intervalField.
    def enterIntervalField(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#intervalField.
    def exitIntervalField(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#normalForm.
    def enterNormalForm(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#normalForm.
    def exitNormalForm(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#type.
    def enterType(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#type.
    def exitType(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#typeParameter.
    def enterTypeParameter(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#typeParameter.
    def exitTypeParameter(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#baseType.
    def enterBaseType(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#baseType.
    def exitBaseType(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#whenClause.
    def enterWhenClause(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#whenClause.
    def exitWhenClause(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#filter.
    def enterFilter(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#filter.
    def exitFilter(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#over.
    def enterOver(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#over.
    def exitOver(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#windowFrame.
    def enterWindowFrame(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#windowFrame.
    def exitWindowFrame(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#unboundedFrame.
    def enterUnboundedFrame(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#unboundedFrame.
    def exitUnboundedFrame(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#currentRowBound.
    def enterCurrentRowBound(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#currentRowBound.
    def exitCurrentRowBound(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#boundedFrame.
    def enterBoundedFrame(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#boundedFrame.
    def exitBoundedFrame(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#explainFormat.
    def enterExplainFormat(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#explainFormat.
    def exitExplainFormat(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#explainType.
    def enterExplainType(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#explainType.
    def exitExplainType(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#isolationLevel.
    def enterIsolationLevel(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#isolationLevel.
    def exitIsolationLevel(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#transactionAccessMode.
    def enterTransactionAccessMode(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#transactionAccessMode.
    def exitTransactionAccessMode(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#readUncommitted.
    def enterReadUncommitted(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#readUncommitted.
    def exitReadUncommitted(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#readCommitted.
    def enterReadCommitted(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#readCommitted.
    def exitReadCommitted(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#repeatableRead.
    def enterRepeatableRead(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#repeatableRead.
    def exitRepeatableRead(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#serializable.
    def enterSerializable(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#serializable.
    def exitSerializable(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#positionalArgument.
    def enterPositionalArgument(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#positionalArgument.
    def exitPositionalArgument(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#namedArgument.
    def enterNamedArgument(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#namedArgument.
    def exitNamedArgument(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#privilege.
    def enterPrivilege(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#privilege.
    def exitPrivilege(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#qualifiedName.
    def enterQualifiedName(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#qualifiedName.
    def exitQualifiedName(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#unquotedIdentifier.
    def enterUnquotedIdentifier(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#unquotedIdentifier.
    def exitUnquotedIdentifier(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#quotedIdentifier.
    def enterQuotedIdentifier(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#quotedIdentifier.
    def exitQuotedIdentifier(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#backQuotedIdentifier.
    def enterBackQuotedIdentifier(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#backQuotedIdentifier.
    def exitBackQuotedIdentifier(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#digitIdentifier.
    def enterDigitIdentifier(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#digitIdentifier.
    def exitDigitIdentifier(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#decimalLiteral.
    def enterDecimalLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#decimalLiteral.
    def exitDecimalLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#doubleLiteral.
    def enterDoubleLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#doubleLiteral.
    def exitDoubleLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#integerLiteral.
    def enterIntegerLiteral(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#integerLiteral.
    def exitIntegerLiteral(self, ctx):
        pass


    # Enter a parse tree produced by SqlBaseParser#nonReserved.
    def enterNonReserved(self, ctx):
        pass

    # Exit a parse tree produced by SqlBaseParser#nonReserved.
    def exitNonReserved(self, ctx):
        pass


